
/* Main benchmark file */

#define INCL_DOSPROCESS
#include <os2.h>

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "types.h"
#include "pmb_bench.h"
#include "pmb_datatype.h"


#define KB 1024
#define MB (1204*1204)
#define d1 (12345678.901*(((double)rand())/RAND_MAX+0.5))
#define d2 (12.345678901*(((double)rand())/RAND_MAX+0.5))
#define MIN_DHRY_TIME 10.0
#define MIN_MEASURE 0.1
#define MARGINAL 1.1

static bool mult;  // true -> runs multiple tests in one thread
static void* p;

extern void PostFin(int onlyupdate);
double dtime(void);
static void EndBench(void);

//extern bool str_has_value[STR_LINES]; // true == the line contains a valid computed value
//extern bool str_is_new[STR_LINES]; // true == value has been changed since last update
//extern double str_value[STR_LINES];

//extern char* str_init[];
extern struct glob_data data;
//static double val_cpufloat[3];
//static double val_cpuint[4];
//static double val_gfx[6];
//static double val_mem[5];

void
DoAll(void* p) {
  mult = true;
  DoAllGraphics(p);
  DoAllCPUInt(p);
  DoAllCPUFloat(p);
//  DoAllDIVE(p);
  DoAllDiskIO(p);
  DoAllMem(p);
  mult = false;
  PostFin(false);
}

void
DoAllCPUFloat(void* p) {
  if (!mult) {
    mult = true;
    DoCPUFloatLinpack(p);
    DoCPUFloatFlops(p);
    DoCPUFloatFFT(p);
    mult = false;
    PostFin(false);
  } else {
    DoCPUFloatLinpack(p);
    DoCPUFloatFlops(p);
    DoCPUFloatFFT(p);
  }
}

void    
DoAllCPUInt(void* p) {
  if (!mult) {
    mult = true;
    DoCPUIntDhry(p);
    DoCPUIntHanoi(p);
    DoCPUIntHeaps(p);
    DoCPUIntSieve(p);
    mult = false;
    PostFin(false);
  } else {
    DoCPUIntDhry(p);
    DoCPUIntHanoi(p);
    DoCPUIntHeaps(p);
    DoCPUIntSieve(p);
  }
}

void    
DoAllDIVE(void* p) {
  if (!mult) {
    mult = true;
    DoDiveRot(p);
    DoDiveVBW(p);
    DoDiveMS11(p);
    DoDiveMS12(p);
    mult = false;
    PostFin(false);
  } else {
    DoDiveRot(p);
    DoDiveVBW(p);
    DoDiveMS11(p);
    DoDiveMS12(p);
  }
}

void    
DoAllDiskIO(void* p) {
  if (!mult) {
    mult = true;
    DoDiskIOAvSeek(p);
    DoDiskIOTransSpeed(p);
    mult = false;
    PostFin(false);
  } else {
    DoDiskIOAvSeek(p);
    DoDiskIOTransSpeed(p);
  }
}

void    
DoAllGraphics(void* p) {
  if (!mult) {
    mult = true;
    DoGfxBlitBlitSS(p);
    DoGfxBlitBlitMS(p);
    DoGfxFillRect(p);
    DoGfxPatFil(p);
    DoGfxVLines(p);
    DoGfxHLines(p);
    DoGfxDLines(p);
    DoGfxTextRender(p);
    mult = false;
    PostFin(false);
  } else { 
    DoGfxBlitBlitSS(p);
    DoGfxBlitBlitMS(p);
    DoGfxFillRect(p);
    DoGfxPatFil(p);
    DoGfxVLines(p);
    DoGfxHLines(p);
    DoGfxDLines(p);
    DoGfxTextRender(p);
  }
}

void    
DoAllMem(void* p) {
  if (!mult) {
    mult = true;
    DoMem5(p);
    DoMem10(p);
    DoMem20(p);
    DoMem40(p);
    DoMem80(p);
    DoMem160(p);
    DoMem320(p);
    DoMem640(p);
    DoMem1280(p);
    DoMemR5(p);
    DoMemR10(p);
    DoMemR20(p);
    DoMemR40(p);
    DoMemR80(p);
    DoMemR160(p);
    DoMemR320(p);
    DoMemR640(p);
    DoMemR1280(p);
    DoMemW5(p);
    DoMemW10(p);
    DoMemW20(p);
    DoMemW40(p);
    DoMemW80(p);
    DoMemW160(p);
    DoMemW320(p);
    DoMemW640(p);
    DoMemW1280(p);
    mult = false;
    PostFin(false);
  } else {
    DoMem5(p);
    DoMem10(p);
    DoMem20(p);
    DoMem40(p);
    DoMem80(p);
    DoMem160(p);
    DoMem320(p);
    DoMem640(p);
    DoMem1280(p);
    DoMemR5(p);
    DoMemR10(p);
    DoMemR20(p);
    DoMemR40(p);
    DoMemR80(p);
    DoMemR160(p);
    DoMemR320(p);
    DoMemR640(p);
    DoMemR1280(p);
    DoMemW5(p);
    DoMemW10(p);
    DoMemW20(p);
    DoMemW40(p);
    DoMemW80(p);
    DoMemW160(p);
    DoMemW320(p);
    DoMemW640(p);
    DoMemW1280(p);
  }
}

extern double pmb_fft(void);

void    
DoCPUFloatFFT(void* p) {
  data.c[comp_cpufloat].datalines[cpufloat_fft].value = pmb_fft();
  EndBench();
}

extern double pmb_flops(void);

void    
DoCPUFloatFlops(void* p) {
  data.c[comp_cpufloat].datalines[cpufloat_flops].value = pmb_flops();
  EndBench();
}

extern double pmb_linpack(void);

void    
DoCPUFloatLinpack(void* p) {
  data.c[comp_cpufloat].datalines[cpufloat_linpack].value = pmb_linpack();
  EndBench();
}

extern double pmb_dhry(int);

void    
DoCPUIntDhry(void* p) {
  double t1, t2, r;
  int loops;
  loops = 1000;
  while (1) {
    t1 = dtime();
    r = pmb_dhry(loops);
    t2 = dtime();
    if (((t2 -t1) < MIN_DHRY_TIME) || (r < 0)) {
      if ((t2 - t1) < MIN_MEASURE) {
        loops = MIN_DHRY_TIME/MIN_MEASURE*loops;
      } else {
        loops = MIN_DHRY_TIME*MARGINAL/(t2-t1)*loops;
      }
    } else {
      break;
    }
  }
  data.c[comp_cpuint].datalines[cpuint_dhrystone].value = r;
  EndBench();
}

extern double pmb_hanoi(void);

void    
DoCPUIntHanoi(void* p) {
  data.c[comp_cpuint].datalines[cpuint_hanoi].value = pmb_hanoi();
  EndBench();
}

extern double pmb_heaps(void);

void    
DoCPUIntHeaps(void* p) {
  data.c[comp_cpuint].datalines[cpuint_heapsort].value = pmb_heaps();
  EndBench();
}

extern double pmb_sieve(void);

void    
DoCPUIntSieve(void* p) {
  data.c[comp_cpuint].datalines[cpuint_sieve].value = pmb_sieve();
  EndBench();
}

extern double pmb_diskio_avseek(int disknr);

void    
DoDiskIOAvSeek(void* p) {
  data.c[comp_disk].datalines[disk_avseek].value = pmb_diskio_avseek(data.selected_disk);
  EndBench();
}

extern double pmb_diskio_transfer(int disknr);

void    
DoDiskIOTransSpeed(void* p) {
  data.c[comp_disk].datalines[disk_transf].value = pmb_diskio_transfer(data.selected_disk);
  EndBench();
}

//extern double pmb_dive_rot(void);

void    
DoDiveRot(void* p) {
  double r;
  r = 1.0; //pmb_dive_rot();
  if (r < 0) {
    r = -1.0;
  }
  data.c[comp_dive].datalines[dive_rotate].value = r;
  EndBench();
}

//extern double pmb_dive_bw(void);

void    
DoDiveVBW(void* p) {
  double r;
  r = 1.0; //pmb_dive_bw();
  if (r < 0) {
    r = -1.0;
  }
  data.c[comp_dive].datalines[dive_videobw].value = r; 
  EndBench();
}

//extern double pmb_dive_ms11(void);

void    
DoDiveMS11(void* p) {
  double r;
  r = 1.0; //pmb_dive_ms11();
  if (r < 0) {
    r = -1.0;
  }
  data.c[comp_dive].datalines[dive_ms_11].value = r;
  EndBench();
}

//extern double pmb_dive_ms12(void);

void    
DoDiveMS12(void* p) {
  double r;
  r = 1.0; //pmb_dive_ms12();
  if (r < 0) {
    r = -1.0;
  }
  data.c[comp_dive].datalines[dive_ms_12].value = r;
  EndBench();
}

//extern double pmb_dive_ms125(void);

void    
DoDiveMS125(void* p) {
  double r;
  r = 1.0; //pmb_dive_ms125();
  if (r < 0) {
    r = -1.0;
  }
  data.c[comp_dive].datalines[dive_ms_125].value = r;
  EndBench();
}

extern double pmb_gfx_bitblitss(void);

void    
DoGfxBlitBlitSS(void* p) {
  data.c[comp_gfx].datalines[gfx_bitblt_SS].value = pmb_gfx_bitblitss();
  EndBench();
}

extern double pmb_gfx_bitblitms(void);

void    
DoGfxBlitBlitMS(void* p) {
  data.c[comp_gfx].datalines[gfx_bitblt_MS].value = pmb_gfx_bitblitms();
  EndBench();
}


extern double pmb_gfx_dlines(void);

void    
DoGfxDLines(void* p) {
  data.c[comp_gfx].datalines[gfx_dlines].value = pmb_gfx_dlines();
  EndBench();
}

extern double pmb_gfx_hlines(void);

void    
DoGfxHLines(void* p) {
  data.c[comp_gfx].datalines[gfx_hlines].value = pmb_gfx_hlines();
  EndBench();
}

extern double pmb_gfx_patrect(void);

void    
DoGfxPatFil(void* p) {
  data.c[comp_gfx].datalines[gfx_patt_fill].value = pmb_gfx_patrect();
  EndBench();
}

extern double pmb_gfx_fillrect(void);

void    
DoGfxFillRect(void* p) {
  data.c[comp_gfx].datalines[gfx_filled_rect].value = pmb_gfx_fillrect();
  EndBench();
}

extern double pmb_gfx_textrender(void);

void    
DoGfxTextRender(void* p) {
  data.c[comp_gfx].datalines[gfx_textrender].value = pmb_gfx_textrender();
  EndBench();
}

extern double pmb_gfx_vlines(void);

void    
DoGfxVLines(void* p) {
  data.c[comp_gfx].datalines[gfx_vlines].value = pmb_gfx_vlines();
  EndBench();
}

extern double pmb_memspeed(s32 bytes);

void    
DoMem5(void* p) {
  data.c[comp_mem].datalines[mem_5].value = pmb_memspeed(5*KB);
  EndBench();
}

void    
DoMem10(void* p) {
  data.c[comp_mem].datalines[mem_10].value = pmb_memspeed(10*KB);
  EndBench();
}

void    
DoMem20(void* p) {
  data.c[comp_mem].datalines[mem_20].value = pmb_memspeed(20*KB);
  EndBench();
}

void    
DoMem40(void* p) {
  data.c[comp_mem].datalines[mem_40].value = pmb_memspeed(40*KB);
  EndBench();
}

void    
DoMem80(void* p) {
  data.c[comp_mem].datalines[mem_80].value = pmb_memspeed(80*KB);
  EndBench();
}

void    
DoMem160(void* p) {
  data.c[comp_mem].datalines[mem_160].value = pmb_memspeed(160*KB);
  EndBench();
}

void    
DoMem320(void* p) {
  data.c[comp_mem].datalines[mem_320].value = pmb_memspeed(320*KB);
  EndBench();
}

void    
DoMem640(void* p) {
  data.c[comp_mem].datalines[mem_640].value = pmb_memspeed(640*KB);
  EndBench();
}

void    
DoMem1280(void* p) {
  data.c[comp_mem].datalines[mem_1280].value = pmb_memspeed(1280*KB);
  EndBench();
}

extern double pmb_memspeedr(s32 bytes);

void    
DoMemR5(void* p) {
  data.c[comp_mem].datalines[memr_5].value = pmb_memspeedr(5*KB);
  EndBench();
}

void    
DoMemR10(void* p) {
  data.c[comp_mem].datalines[memr_10].value = pmb_memspeedr(10*KB);
  EndBench();
}

void    
DoMemR20(void* p) {
  data.c[comp_mem].datalines[memr_20].value = pmb_memspeedr(20*KB);
  EndBench();
}

void    
DoMemR40(void* p) {
  data.c[comp_mem].datalines[memr_40].value = pmb_memspeedr(40*KB);
  EndBench();
}

void    
DoMemR80(void* p) {
  data.c[comp_mem].datalines[memr_80].value = pmb_memspeedr(80*KB);
  EndBench();
}

void    
DoMemR160(void* p) {
  data.c[comp_mem].datalines[memr_160].value = pmb_memspeedr(160*KB);
  EndBench();
}

void    
DoMemR320(void* p) {
  data.c[comp_mem].datalines[memr_320].value = pmb_memspeedr(320*KB);
  EndBench();
}

void    
DoMemR640(void* p) {
  data.c[comp_mem].datalines[memr_640].value = pmb_memspeedr(640*KB);
  EndBench();
}

void    
DoMemR1280(void* p) {
  data.c[comp_mem].datalines[memr_1280].value = pmb_memspeedr(1280*KB);
  EndBench();
}

extern double pmb_memspeedw(s32 bytes);

void    
DoMemW5(void* p) {
  data.c[comp_mem].datalines[memw_5].value = pmb_memspeedw(5*KB);
  EndBench();
}

void    
DoMemW10(void* p) {
  data.c[comp_mem].datalines[memw_10].value = pmb_memspeedw(10*KB);
  EndBench();
}

void    
DoMemW20(void* p) {
  data.c[comp_mem].datalines[memw_20].value = pmb_memspeedw(20*KB);
  EndBench();
}

void    
DoMemW40(void* p) {
  data.c[comp_mem].datalines[memw_40].value = pmb_memspeedw(40*KB);
  EndBench();
}

void    
DoMemW80(void* p) {
  data.c[comp_mem].datalines[memw_80].value = pmb_memspeedw(80*KB);
  EndBench();
}

void    
DoMemW160(void* p) {
  data.c[comp_mem].datalines[memw_160].value = pmb_memspeedw(160*KB);
  EndBench();
}

void    
DoMemW320(void* p) {
  data.c[comp_mem].datalines[memw_320].value = pmb_memspeedw(320*KB);
  EndBench();
}

void    
DoMemW640(void* p) {
  data.c[comp_mem].datalines[memw_640].value = pmb_memspeedw(640*KB);
  EndBench();
}

void    
DoMemW1280(void* p) {
  data.c[comp_mem].datalines[memw_1280].value = pmb_memspeedw(1280*KB);
  EndBench();
}

void    
DoRef1Info(void) {
}

void    
DoRef1Load(void) {
}

void    
DoRef2Info(void) {
}

void    
DoRef2Load(void) {
}

double    
CalcGfxAv(void) {
  return (
          data.c[comp_gfx].datalines[gfx_bitblt_SS].value * 10 + 
          data.c[comp_gfx].datalines[gfx_bitblt_MS].value * 8 + 
          data.c[comp_gfx].datalines[gfx_filled_rect].value * 7 + 
          data.c[comp_gfx].datalines[gfx_patt_fill].value * 4 + 
          data.c[comp_gfx].datalines[gfx_vlines].value * 5 + 
          data.c[comp_gfx].datalines[gfx_hlines].value * 5 + 
          data.c[comp_gfx].datalines[gfx_dlines].value * 4 + 
          data.c[comp_gfx].datalines[gfx_textrender].value * 8
         ) / 51 / 1.0e6;
//  return (val_gfx[0]*10 + val_gfx[1].value*5 + val_gfx[2]*3 + val_gfx[3]*3 + 
//          val_gfx[4]*3 + val_gfx[5]*8)/32/1.0e6;
}

double    
CalcCPUIntAv(void) {
  return (
          data.c[comp_cpuint].datalines[cpuint_dhrystone].value * 12 + 
          data.c[comp_cpuint].datalines[cpuint_hanoi].value * 5 * 1.0e6+ 
          data.c[comp_cpuint].datalines[cpuint_heapsort].value * 6 + 
          data.c[comp_cpuint].datalines[cpuint_sieve].value * 6 
         ) / 29 / 1.0e6;
//  return (val_cpuint[0]*12 + val_cpuint[1]*5 + val_cpuint[2]*6 + val_cpuint[3]*6)/29/1.0e6;
}

double    
CalcCPUFloatAv(void) {
  return (
          data.c[comp_cpufloat].datalines[cpufloat_linpack].value * 10 * 1000 + 
          data.c[comp_cpufloat].datalines[cpufloat_flops].value * 20 + 
          data.c[comp_cpufloat].datalines[cpufloat_fft].value * 9 * 1.0e6 
         ) / 39 / 1.0e6;
//  return (val_cpufloat[0]*10 + val_cpufloat[1]*20 + val_cpufloat[2]*9)/39/1.0e6;
}

double    
CalcMemAv(void) {
  return (
          data.c[comp_mem].datalines[mem_5].value * 7 + 
          data.c[comp_mem].datalines[mem_10].value * 8 + 
          data.c[comp_mem].datalines[mem_20].value * 7 + 
          data.c[comp_mem].datalines[mem_40].value * 7 + 
          data.c[comp_mem].datalines[mem_80].value * 6 + 
          data.c[comp_mem].datalines[mem_160].value * 5 + 
          data.c[comp_mem].datalines[mem_320].value * 4 + 
          data.c[comp_mem].datalines[mem_640].value * 4 + 
          data.c[comp_mem].datalines[mem_1280].value * 3 +
          data.c[comp_mem].datalines[memr_5].value * 7 + 
          data.c[comp_mem].datalines[memr_10].value * 8 + 
          data.c[comp_mem].datalines[memr_20].value * 7 + 
          data.c[comp_mem].datalines[memr_40].value * 7 + 
          data.c[comp_mem].datalines[memr_80].value * 6 + 
          data.c[comp_mem].datalines[memr_160].value * 5 + 
          data.c[comp_mem].datalines[memr_320].value * 4 + 
          data.c[comp_mem].datalines[memr_640].value * 4 + 
          data.c[comp_mem].datalines[memr_1280].value * 3 +
          data.c[comp_mem].datalines[memw_5].value * 7 + 
          data.c[comp_mem].datalines[memw_10].value * 8 + 
          data.c[comp_mem].datalines[memw_20].value * 7 + 
          data.c[comp_mem].datalines[memw_40].value * 7 + 
          data.c[comp_mem].datalines[memw_80].value * 6 + 
          data.c[comp_mem].datalines[memw_160].value * 5 + 
          data.c[comp_mem].datalines[memw_320].value * 4 + 
          data.c[comp_mem].datalines[memw_640].value * 4 + 
          data.c[comp_mem].datalines[memw_1280].value * 3
         ) / (51*3.0) / MB;
//  return (val_mem[0]*10 + val_mem[1]*10 + val_mem[2]*15 + val_mem[3]*10 + val_mem[4]*8)/53/MB;
}

double    
CalcDIVEAv(void) {
  return (
          data.c[comp_dive].datalines[dive_videobw].value * 10 + 
          data.c[comp_dive].datalines[dive_rotate].value * 3 * 1.0e6+ 
          data.c[comp_dive].datalines[dive_ms_11].value * 10* 1.0e6 + 
          data.c[comp_dive].datalines[dive_ms_12].value * 10* 1.0e6 +
          data.c[comp_dive].datalines[dive_ms_125].value * 10 * 1.0e6
         ) / 43 / 1.0e6;
}

double    
CalcDiskIOAv(void) {
  return (
          data.c[comp_disk].datalines[disk_avseek].value * 5 * 1.0e9 / 20.0 + 
          data.c[comp_disk].datalines[disk_transf].value * 5 
         ) / 10 / 1.0e6;
}




double dtime(void)
{
    return ((double)clock())/CLOCKS_PER_SEC;
}

double rtime(void)
{
    return ((double)clock())/CLOCKS_PER_SEC;
}

void log(char* s) {
  FILE* f;
  f = fopen("sb_error.log", "a+");
  if (!f) {
    exit(234);
  }
  fprintf(f, "%s\n", s);
  fclose(f);
}

static void
EndBench(void) {
  if (!mult) {
    PostFin(true);
    PostFin(false);
  } else {
    PostFin(true);
    DosSleep(500); // wait for the window thread to update the screen
  }
}

